;;; ksc5601.el --- set up for KSC5601 for Korean ttys

;; Copyright (C) 1994 Deoktae Kim

;; Author: Deoktae Kim <dtkim@camars.kaist.ac.kr>
;; Maintainer: Deoktae Kim <dtkim@camars.kaist.ac.kr>
;; Created: 10 Jan 1994
;; Version: 1.3 (Tested under GNU Emacs 19.22.5)
;; Last Modified: Mon Jan 10 09:23:35 KST 1994
;; Keywords: i18n

;; This file is part of GNU Emacs.

;; GNU Emacs is free software; you can redistribute it and/or modify
;; it under the terms of the GNU General Public License as published by
;; the Free Software Foundation; either version 2, or (at your option)
;; any later version.

;; GNU Emacs is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;; GNU General Public License for more details.

;; You should have received a copy of the GNU General Public License
;; along with GNU Emacs; see the file COPYING.  If not, write to
;; the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

;;; Commentary:

;; Input, syntax, and display for KSC5601 codes are supported.
; Following code should be included in site-start.el or your .emacs file.
;(if (and (null window-system) (null noninteractive)) 
;    (progn
;      (load-library "ksc5601")
;      (standard-display-ksc5601 t)
;      (substitute-key-definition
;       'backward-delete-char-untabify
;       'ksc5601-backward-delete-char-untabify
;       lisp-interaction-mode-map)))

;;; Known Bugs:

;; - Since there is no difference between English and Korean syntax
;;      class, two adjacent English word and Korean word are treated
;;      as just one word.
;; - Failure of handling stale file.
;; - Insertion of character at begging of buffer need to be redrawn twice
;;   in Hanterm (seems to be Hanterm bug)
;; - In, overwrite-mode, Korean characters need to be redrawn by user in
;;   Hanterm (seems to be Hanterm bug)

;;; What to do:

;; - Support for picture mode.
;; - Revison of display toggling behavior.
;; - Support for isearch. (e.g., isearch-mode-map)
;; - `C-x =' should show hangul or at least do not show half of hangul like `M-3'.
;; - Support for transpose-chars
;; - Consecutive updates should be merged in undo list.
;; - Inspect the behavior when ispell or abbreviation mode is executed.

;;; Alogrithms:
;; - Movement command algorithm::
;;    Inspect the target position and if the display glyph is non-nil and in range (161, 254),
;;    then search backward some char in the not range, and then 

;;; Code:

(set-input-mode (car (current-input-mode))
		(nth 1 (current-input-mode))
		0)

;; For making syntax class of Korean letters as `words'.
(defun ksc5601-to-syntax (syntaxtab)
  (let ((i 161))
    (while (<= i 254)
      (modify-syntax-entry i "w   " syntaxtab)
      (setq i (1+ i)))))

(ksc5601-to-syntax (standard-syntax-table))
(ksc5601-to-syntax emacs-lisp-mode-syntax-table)
(defun standard-display-ksc5601 (arg)
  "Toggle display of Korean characters encoded with KSC5601 by default.
When enabled, characters in the range of 161 to 254 display not
as octal escapes, but as Korean characters.
With prefix argument, enable KSC5601 character display iff ARG is positive."
  (interactive "P")
  (if (or (< (prefix-numeric-value arg) 0)
	  (and (null arg)
	       (vectorp standard-display-table)
	       (>= (length standard-display-table) 162)
	       (equal (aref standard-display-table 161) [161])))
      (standard-display-default 161 254)
    (standard-display-8bit 161 254)))

(defun buffer-display-ksc5601 (arg)
  "Toggle display of Korean characters encoded with KSC5601 of selected buffer.
When enabled, characters in the range of 161 to 254 display not
as octal escapes, but as Korean characters.
With prefix argument, enable KSC5601 character display iff ARG is positive."
  (interactive "P")
  (let ((off-flag (or (< (prefix-numeric-value arg) 0)
		      (and (null arg)
			   (vectorp buffer-display-table)
			   (>= (length buffer-display-table) 162)
			   (equal (aref buffer-display-table 161) [161]))))
	(l 161))
    (or buffer-display-table
	(setq buffer-display-table (make-vector 261 nil)))
    (while (<= l 254)
      (aset buffer-display-table l (if off-flag nil (vector l)))
      (setq l (1+ l)))))

(defun window-display-ksc5601 (arg)
  "Toggle display of Korean characters encoded with KSC5601 of selected window.
When enabled, characters in the range of 161 to 254 display not
as octal escapes, but as Korean characters.
With prefix argument, enable KSC5601 character display iff ARG is positive."
  (interactive "P")
  (let* ((disp-tab (window-display-table (selected-window)))
	 (off-flag (or (< (prefix-numeric-value arg) 0)
		       (and (null arg)
			    (vectorp disp-tab)
			    (>= (length disp-tab) 162)
			    (equal (aref disp-tab 161) [161]))))
	 (l 161))
    (or disp-tab
	(setq disp-tab (make-vector 261 nil)))
    (while (<= l 254)
      (aset disp-tab l (if off-flag nil (vector l)))
      (setq l (1+ l)))
    (set-window-display-table (selected-window) disp-tab)))

(defun display-ksc5601-p ()
  (let ((disptab
	 (or (window-display-table (selected-window))
	     buffer-display-table
	     standard-display-table)))
    (if disptab
	(equal (aref disptab 161) [161]))))

(substitute-key-definition
 'forward-char 'ksc5601-forward-char global-map)
(substitute-key-definition
 'backward-char 'ksc5601-backward-char global-map)
(substitute-key-definition
 'previous-line 'ksc5601-previous-line global-map)
(substitute-key-definition
 'next-line 'ksc5601-next-line global-map)
(substitute-key-definition
 'delete-char 'ksc5601-delete-char global-map)
(substitute-key-definition
 'delete-backward-char 'ksc5601-delete-backward-char global-map)
(substitute-key-definition
 'self-insert-command 'ksc5601-self-insert-command global-map)
(substitute-key-definition
 'backward-delete-char-untabify 'ksc5601-backward-delete-char-untabify global-map)


(defun ksc5601-forward-char (&optional n)
  "Move point right ARG characters (left if ARG negative). Adapted to handle KSC5601.
On reaching end of buffer, stop and signal error."
  (interactive "p")
  (let ((target-pos (+ (point) n)))
    (if (or (not (display-ksc5601-p))
	    (<= target-pos (point-min))
	    (>= target-pos (point-max))
	    (and (= n 1) (>= (following-char) 161) (<= (following-char) 254) (setq n 2))
	    (< (char-after target-pos) 161)
	    (> (char-after target-pos) 254))
	(forward-char n)
      (forward-char n)
      (forward-char (logand (- (skip-chars-backward "\241-\376" (point-min))) -2)))))

(defun ksc5601-backward-char (&optional n)
  "Move point left ARG characters (right if ARG negative). Adapted to handle KSC5601.
On attempt to pass beginning or end of buffer, stop and signal error."
  (interactive "p")
  (ksc5601-forward-char (- n)))

(defun ksc5601-goto-char (&optional n)
  "Set point to POSITION, a number or marker. Adapted to handle KSC5601.
Beginning of buffer is n (point-min), end is (point-max).
Adjust point to the first position of KSC5601 characters."
  (interactive "p")
  (goto-char n)
  (if (and (display-ksc5601-p)
	   (char-after n)
	   (>= (char-after n) 161)
	   (<= (char-after n) 254))
      (forward-char (logand (- (skip-chars-backward "\241-\376" (point-min))) -2))))

(defun ksc5601-next-line (arg)
  "Move cursor vertically down ARG lines. Adapted to handle KSC5601.
If there is no character in the target line exactly under the current column,
the cursor is positioned after the character in that line which spans this
column, or at the end of the line if it is not long enough.
If there is no line in the buffer after this one, behavior depends on the
value of next-line-add-newlines.  If non-nil, a newline character is inserted
to create a line and the cursor moves to that line, otherwise the cursor is
moved to the end of the buffer (if already at the end of the buffer, an error
is signaled).

The command \\[set-goal-column] can be used to create
a semipermanent goal column to which this command always moves.
Then it does not try to move vertically.  This goal column is stored
in `goal-column', which is nil when there is none.

If you are thinking of using this in a Lisp program, consider
using `forward-line' instead.  It is usually easier to use
and more reliable (no dependence on goal column, etc.)."
  (interactive "p")
  (let ((opoint (point)))
    (if next-line-add-newlines
	(if (/= arg 1)
	    (ksc5601-line-move arg)
	  (forward-line 1)
	  (if (or (= opoint (point)) (not (eq (preceding-char) ?\n)))
	      (insert ?\n)
	    (goto-char opoint)
	    (ksc5601-line-move arg)))
      (if (eobp)
	  (signal 'end-of-buffer nil))
      (ksc5601-line-move arg)
      (if (= opoint (point))
	  (end-of-line))))
  nil)

(defun ksc5601-previous-line (arg)
  "Move cursor vertically up ARG lines. Adapted to handle KSC5601.
If there is no character in the target line exactly over the current column,
the cursor is positioned after the character in that line which spans this
column, or at the end of the line if it is not long enough.

The command \\[set-goal-column] can be used to create
a semipermanent goal column to which this command always moves.
Then it does not try to move vertically.

If you are thinking of using this in a Lisp program, consider using
`forward-line' with a negative argument instead.  It is usually easier
to use and more reliable (no dependence on goal column, etc.)."
  (interactive "p")
  (ksc5601-line-move (- arg))
  nil)

(defun ksc5601-line-move (arg)
  (if (not (or (eq last-command 'ksc5601-next-line)
	       (eq last-command 'ksc5601-previous-line)
	       (eq last-command 'next-line)
	       (eq last-command 'previous-line)))
      (setq temporary-goal-column
	    (if (and track-eol (eolp)
		     ;; Don't count beg of empty line as end of line
		     ;; unless we just did explicit end-of-line.
		     (or (not (bolp)) (eq last-command 'end-of-line)))
		9999
	      (current-column))))
  (if (not (integerp selective-display))
      (forward-line arg)
    ;; Move by arg lines, but ignore invisible ones.
    (while (> arg 0)
      (vertical-motion 1)
      (forward-char -1)
      (forward-line 1)
      (setq arg (1- arg)))
    (while (< arg 0)
      (vertical-motion -1)
      (beginning-of-line)
      (setq arg (1+ arg))))
  (ksc5601-move-to-column (or goal-column temporary-goal-column))
  nil)

(defun ksc5601-move-to-column (column &optional force)
  (move-to-column column force)
  (forward-char (logand (- (skip-chars-backward "\241-\376" (point-min))) -2)))

(defun ksc5601-delete-char (n &optional killflag)
  "Delete the following ARG characters (previous, with negative arg). Adapted to handle KSC5601.
Optional second arg KILLFLAG non-nil means kill instead (save in kill ring).
Interactively, ARG Is the prefix arg, and KILLFLAG is set if
ARG was explicitly specified."
  (interactive "p\nP")
  (let ((end-pos (+ (point) n)))
    (if (or (not (display-ksc5601-p))
	    (not (integerp n))
	    (<= end-pos (point-min))
	    (>= end-pos (point-max))
	    (and (= n 1) (>= (following-char) 161) (<= (following-char) 254) (setq n 2))
	    (< (char-after end-pos) 161)
	    (> (char-after end-pos) 254))
	(delete-char n killflag)
      (let ((align))
	(save-excursion
	  (forward-char n)
	  (setq align (logand (- (skip-chars-backward "\241-\376" (point-min))) 1)))
	(if (= align 0)
	    (delete-char n killflag)
	  (if (>= n 0)
	      (delete-char (+ n 1) killflag)
	    (delete-char (- n 1) killflag)))))))

(defun ksc5601-delete-backward-char (n &optional killflag)
  "Delete the previous ARG characters (following, with negative ARG). Adapted to handle KSC5601.
Optional second arg KILLFLAG non-nil means kill instead (save in kill ring).
Interactively, ARG is the prefix arg, and KILLFLAG is set if
ARG was explicitly specified."

  (interactive "p\nP")
  (ksc5601-delete-char (- n) killflag))

;(fset 'org-self-insert-command (symbol-function 'self-insert-command))
(defun ksc5601-self-insert-command (n)
  "Insert the character you type. Adapted to handle KSC5601.
Whichever character you type to run this command is inserted.
When you insert KSC5601 code in that display mode, n must not be
greater than 1"
  (interactive "p")
  (if (or (not (interactive-p))
	  (not overwrite-mode)
	  (not (display-ksc5601-p))
	  (not (integerp n))
	  (<= n 0))
      (progn (self-insert-command n)
	     (and (eq last-command 'ksc5601-self-insert-command)
		  (null prefix-arg)
		  (boundp last-prefix-arg)
		  (null last-prefix-arg)
		  (let* ((prev-undo (cdr (cdr buffer-undo-list)))
		        (prev-diff (- (cdr (car prev-undo)) (car (car prev-undo)))))
		       (and (< prev-diff 20)
			    (or (/= prev-diff 19)
				(eq 0 (logand (save-excursion (skip-chars-backward "\241-\376" (point-min))) 1)))
			    (setq buffer-undo-list
				  (cons (cons (car (car prev-undo))
					      (1+ (cdr (car prev-undo))))
					(cdr prev-undo)))))))
    (self-insert-command n)
    (if (and (>= (following-char) 161)
	     (<= (following-char) 254)
	     (eq 1 (logand (save-excursion (skip-chars-forward "\241-\376" (point-max))) 1)))
	(progn (delete-char 1)
	       (insert ?\ )
	       (forward-char -1))))
  (setq last-prefix-arg prefix-arg)
  nil)

(defun ksc5601-backward-delete-char-untabify (arg &optional killp)
  "Delete characters backward, changing tabs into spaces. Adapted to handle KSC5601.
Delete ARG chars, and kill (save in kill ring) if KILLP is non-nil.
Interactively, ARG is the prefix arg (default 1)
and KILLP is t if prefix arg is was specified."
  (interactive "*p\nP")
  (backward-delete-char-untabify arg killp)
  (if (display-ksc5601-p)
      (if (and (> arg 0)
	       (>= (preceding-char) 161)
	       (<= (preceding-char) 254)
	       (save-excursion (eq 1 (logand (- (skip-chars-backward "\241-\376" (point-min))) 1))))
	  (backward-delete-char-untabify 1 killp)
	(if (and (< arg 0)
		 (>= (following-char) 161)
		 (<= (following-char) 254)
		 (save-excursion (eq 1 (logand (skip-chars-forward "\241-\376" (point-max))) 1)))
	    (backward-delete-char-untabify -1 killp)))))


;;; Support for sending mail in KSC5601
;; You need to install /usr/local/bin/ks2sdn, /usr/local/bin/sdn2ks,
;;   and(hopefully) /bin/mail.
;; You can ftp them from cair:/hangul/mail/hmail1.22.
;; If your system administrator installed /bin/mail from hmail1.22
;;   everything should work well.
;; If not, request to install them to your system administrator,
;;   or you can set your .forward file to include things like following:
;;   "|/usr/local/bin/sdn2ks|/usr/local/bin/filter"
;; If all above are not appropriate choices, you can use the command
;;   `shell-command-on-region' with argument `sdn2ks' on mail messages to get
;;   proper form in KSC5601.

;(setq send-mail-function 'ksc5601-sendmail-send-it)

(defun ksc5601-sendmail-send-it ()
  (let ((errbuf (if mail-interactive
		    (generate-new-buffer " sendmail errors")
		  0))
	(tembuf (generate-new-buffer " sendmail temp"))
	(case-fold-search nil)
	delimline
	(mailbuf (current-buffer)))
    (unwind-protect
	(save-excursion
	  (set-buffer tembuf)
	  (erase-buffer)
	  (insert-buffer-substring mailbuf)
	  (goto-char (point-max))
	  ;; require one newline at the end.
	  (or (= (preceding-char) ?\n)
	      (insert ?\n))
	  ;; Change header-delimiter to be what sendmail expects.
	  (goto-char (point-min))
	  (re-search-forward
	    (concat "^" (regexp-quote mail-header-separator) "\n"))
	  (replace-match "\n")
	  (backward-char 1)
	  (setq delimline (point-marker))
	  (if mail-aliases
	      (expand-mail-aliases (point-min) delimline))
	  (goto-char (point-min))
	  ;; ignore any blank lines in the header
	  (while (and (re-search-forward "\n\n\n*" delimline t)
		      (< (point) delimline))
	    (replace-match "\n"))
	  (let ((case-fold-search t))
	    (goto-char (point-min))
	    (if (re-search-forward "^Sender:" delimline t)
		(error "Sender may not be specified."))
	    ;; Find and handle any FCC fields.
	    (goto-char (point-min))
	    (if (re-search-forward "^FCC:" delimline t)
		(mail-do-fcc delimline))
;;; Apparently this causes a duplicate Sender.
;;;	    ;; If the From is different than current user, insert Sender.
;;;	    (goto-char (point-min))
;;;	    (and (re-search-forward "^From:"  delimline t)
;;;		 (progn
;;;		   (require 'mail-utils)
;;;		   (not (string-equal
;;;			 (mail-strip-quoted-names
;;;			  (save-restriction
;;;			    (narrow-to-region (point-min) delimline)
;;;			    (mail-fetch-field "From")))
;;;			 (user-login-name))))
;;;		 (progn
;;;		   (forward-line 1)
;;;		   (insert "Sender: " (user-login-name) "\n")))
	    ;; "S:" is an abbreviation for "Subject:".
	    (goto-char (point-min))
	    (if (re-search-forward "^S:" delimline t)
		(replace-match "Subject:"))
	    ;; Don't send out a blank subject line
	    (goto-char (point-min))
	    (if (re-search-forward "^Subject:[ \t]*\n" delimline t)
		(replace-match ""))
	    (if mail-interactive
		(save-excursion
		  (set-buffer errbuf)
		  (erase-buffer))))

;;;; This part is modified to below of it to handle sending mail in KSC5601 by dtkim@camars.kaist.ac.kr.
;;;	  (apply 'call-process-region
;;;		 (append (list (point-min) (point-max)
;;;			       (if (boundp 'sendmail-program)
;;;				   sendmail-program
;;;				 "/usr/lib/sendmail")
;;;			       nil errbuf nil
;;;			       "-oi" "-t")
;;;			 ;; Always specify who from,
;;;			 ;; since some systems have brvoken sendmails.
;;;			 (list "-f" (user-login-name))
;;;;;;			 ;; Don't say "from root" if running under su.
;;;;;;			 (and (equal (user-real-login-name) "root")
;;;;;;			      (list "-f" (user-login-name)))
;;;			 (and mail-alias-file
;;;			      (list (concat "-oA" mail-alias-file)))
;;;			 ;; These mean "report errors by mail"
;;;			 ;; and "deliver in background".
;;;			 (if (null mail-interactive) '("-oem" "-odb"))))

;;;; This part is to handle sending mail in KSC5601 by dtkim@camars.kaist.ac.kr
	  (apply 'call-process-region
		 (list (point-min) (point-max)
		       "/bin/sh"
		       nil errbuf nil
		       "-c"
		       (concat "/usr/local/bin/ks2sdn |"
			       (if (boundp 'sendmail-program)
				   sendmail-program
				 " /usr/lib/sendmail")
			       " -oi" " -t"
			       ;; Always specify who from,
			       ;; since some systems have broken sendmails.
			       (concat " -f " (user-login-name))
;;;			 ;; Don't say "from root" if running under su.
;;;			 (and (equal (user-real-login-name) "root")
;;;			      (concat " -f " (user-login-name))
			       (and mail-alias-file
				    (concat " -oA " mail-alias-file))
			       ;; These mean "report errors by mail"
			       ;; and "deliver in background".
			       (if (null mail-interactive) (concat " -oem " " -odb ")))))
	  (if mail-interactive
	      (save-excursion
		(set-buffer errbuf)
		(goto-char (point-min))
		(while (re-search-forward "\n\n* *" nil t)
		  (replace-match "; "))
		(if (not (zerop (buffer-size)))
		    (error "Sending...failed to %s"
			   (buffer-substring (point-min) (point-max)))))))
      (kill-buffer tembuf)
      (if (bufferp errbuf)
	  (kill-buffer errbuf)))))

(provide 'ksc5601)

;;; ksc5601.el ends here
